#include <stdio.h> 
#include <string.h> 
#include <cmqc.h> 
#include <threads.h> 
 
#define MAX_QUEUES 20 
#define NUM_QUEUES_DEFAULT 3 
#define Q1_DEFAULT "L1.Q" 
#define Q2_DEFAULT "L2.Q" 
#define Q3_DEFAULT "L3.Q" 
#define IN
#define OUT
 
#define SLEEP_TIME 1 

#if 1 
	#define XML_MESSAGE \
	"<outer1><tag1>data1</tag1><tag2>data2</tag2><tag3>data3</tag3><reply>data4</reply></outer1>" 
#else
#if 0
	#define XML_MESSAGE \
	"<?xml version=\"1.0\" encoding=\"UTF-8\"?><outer1><tag1>data1</tag1><tag2>data2</tag2><tag3>data3</tag3><reply>data4</reply></outer1>" 
#else
	#define XML_MESSAGE \
	"<?xml version='1.0' encoding='UTF-8'?><outer1><tag1 attr1='fred'>data1</tag1><tag2>data2</tag2><tag3>data3</tag3><reply>data4</reply></outer1>" 
#endif
#endif

FUNCTION_RETURN_TYPE processMessages(void *dataP); 
 
typedef struct threadData { 
	char *queueName; 
	bool_t *threadInitialisedP; 
} threadData_t; 
 
#if defined(WIN32) 
    #define QM_DEFAULT "NT_QM" 
#elif defined(_AIX) 
    #define QM_DEFAULT "SEAIX_QM" 
#elif defined(_SOLARIS) 
    #define QM_DEFAULT "SUN_QM" 
#elif defined(_HPUX)
    #define QM_DEFAULT "HP_QM"
#endif 
 
char *qmName; 
char *inputQ[MAX_QUEUES]; 
int qCount = 0; 
 
threadMutex_t workMutex; 
threadCondVar_t workCondVar, 
		        syncCondVar; 
int main(int argc, char **argv) 
{ 
	bool_t threadInitialised; 
	int retcode, 
		i, 
		j, 
		argCount, 
		numQueues; 
	threadHandle_t threadArray[MAX_QUEUES], 
		           *handleP; 
	threadData_t threadInitData; 
 
 
	/* display usage template if requested */ 
    if (argc == 2 && argv[1][0] == '?') 
	{ 
		printf("Usage: %s < <queueManagerName> <queueName1> <queueName2>...<queueName n>>\n", 
			    argv[0]); 
		printf("Usage: Maximum number of queues that can be specified is 20\n"); 
		return(0); 
	} 
 
    if (argc > 1) 
    { 
		if (argc > 22)  /* progName + qmName + 20 queueNames */ 
		{ 
			fprintf(stderr, "Too many queue names, max is 20\n"); 
			return(1); 
		} 
 
        /* read in queue manager name */ 
		qmName = argv[1]; 
 
		/* now read the queue name(s) */ 
		numQueues = 0; 
		for (j=0, argCount=3; argCount<=argc; j++, argCount++) 
		{ 
			inputQ[j] = argv[argCount-1]; 
			numQueues++; 
		} 
 
    } 
    else 
    { 
		qmName = QM_DEFAULT; 
		inputQ[0] = Q1_DEFAULT; 
		inputQ[1] = Q2_DEFAULT; 
		inputQ[2] = Q3_DEFAULT; 
        numQueues = NUM_QUEUES_DEFAULT; 
    } 
 
	printf("Queue manager: %s\n", qmName); 
	for (i=0; i<numQueues; i++) 
	{ 
		printf("Queue name[%d]: %s\n", i+1, inputQ[i]); 
	} 
 
	/* create synchronisation objects */ 
	if (threadCreateMutex(FALSE,  /* create in unlocked state */ 
						  &workMutex) != THREAD_OK) 
	{ 
	    printf("Failed to create workMutex\n"); 
		return(1); 
	} 
 
	if (threadCreateCondVar(workMutex, 
							&workCondVar) != THREAD_OK) 
	{ 
	    printf("Failed to create workCondVar\n"); 
		return(1); 
	} 
 
	if (threadCreateCondVar(workMutex, 
							&syncCondVar) != THREAD_OK) 
	{ 
	    printf("Failed to create syncCondVar\n"); 
		return(1); 
	} 
 
	/* set up the pool of worker threads */ 
    threadMutexLock(workMutex); 
	for (i=0; i<numQueues; i++) 
	{ 
		handleP = threadArray+i; 
		threadInitialised = FALSE; 
		threadInitData.queueName = inputQ[i]; 
		threadInitData.threadInitialisedP = &threadInitialised; 
 
		if (threadCreateThread((threadFunction_t)processMessages, 
						       (threadParameter_t)&threadInitData, 
						       handleP) != THREAD_OK) 
		{ 
			fprintf(stderr, "Error creating worker thread\n"); 
			return(1); 
		} 
 
	    retcode = threadCondTimedWait(syncCondVar, 10000); 
		if (retcode == THREAD_TIMED_OUT) 
		{ 
			fprintf(stderr, "parent thread wait timed out - exiting\n"); 
			return(1); 
		} 
 
		if (retcode != THREAD_OK) 
		{ 
			fprintf(stderr, "threadCondTimedWait failed, rc=%d\n", retcode); 
			return(1); 
		} 
 
		if (!threadInitialised) 
		{ 
			fprintf(stderr, "worker %d thread failed to initialise cleanly\n", 
                            threadGetThreadId(*handleP)); 
			return(1); 
		} 
		else 
		{ 
			printf("worker thread %d initialised ok\n", 
                            threadGetThreadId(*handleP)); 
		} 
	} 
 
    threadMutexUnlock(workMutex); 
 
	/* wait for worker threads to die...never happens till kill is issued */ 
	for (i=0; i<numQueues; i++) 
	{ 
		printf("waiting for thread %d to terminate...\n", threadGetThreadId(*(threadArray+i))); 
		retcode = threadJoin(*(threadArray+i)); 
	} 
 
    return(0); 
} 

void dispHex(IN char *inputString, 
			 IN int inputStringlength, 
			 OUT char *hexString);

 
FUNCTION_RETURN_TYPE processMessages(void *dataP) 
{ 
	MQHCONN hConn; 
	MQHOBJ requestQHandle; 
    MQLONG   openOptions;               /* MQOPEN options                */ 
    MQOD od, 
		 odDefault = {MQOD_DEFAULT};           /* Object Descriptor */ 
    MQPMO pmo = {MQPMO_DEFAULT};        /* put message options */ 
    MQGMO gmo = {MQGMO_DEFAULT};        /* get message options */ 
	MQMD md, 
		 mdDefault = {MQMD_DEFAULT};           /* Message Descriptor */ 
	MQLONG compCode, 
		   reason; 
	threadData_t *threadInitDataP = (threadData_t *)dataP; 
	char myQueueName[100], 
		 messageBuffer[2000], 
		 requestId[MQ_MSG_ID_LENGTH]; 
	int messageBufferLength, 
		messageLength; 
    int threadId, 
		keepLooping; 
	char hexBuffer[MQ_MSG_ID_LENGTH*2 + 1];

    threadMutexLock(workMutex); 
 
	threadId = threadGetCurrentThreadId(); 
 
	printf("Thread %d: before call to MQCONN\n", threadId); 
 
	/* connect to queue manager, assume default */ 
	MQCONN(qmName, 
		   &hConn, 
		   &compCode, 
		   &reason); 
	if (reason != MQRC_NONE) 
	{ 
		printf("Thread %d: MQCONN failed with compcode=%d, reason=%d\n", threadId, 
			   compCode, reason); 
        threadCondSignal(syncCondVar); 
        threadMutexUnlock(workMutex); 
		return((FUNCTION_RETURN_TYPE)1); 
	} 
 
	printf("Thread %d: Connection handle returned is %.8x\n", threadId, hConn); 
 
	/* open the specified input queue */ 
	memcpy(&od, &odDefault, sizeof(od)); 
	strcpy(myQueueName, threadInitDataP->queueName); 
	printf("Thread %d: About to open queue '%s'\n", threadId, myQueueName); 
	strncpy(od.ObjectName, 
			myQueueName, 
			(size_t)MQ_Q_NAME_LENGTH); 
 
	/* set the open mode */ 
	openOptions = MQOO_INPUT_AS_Q_DEF +  
		          MQOO_FAIL_IF_QUIESCING; /* open queue for output */ 
 
	MQOPEN(hConn, 
		   &od, 
		   openOptions, 
		   &requestQHandle, 
		   &compCode, 
		   &reason); 
	if (reason != MQRC_NONE) 
	{ 
		printf("Thread %d: MQOPEN for input queue '%s' failed with compcode=%d, reason=%d\n", 
			   threadId, 
			   myQueueName, compCode, reason); 
        threadCondSignal(syncCondVar); 
        threadMutexUnlock(workMutex); 
		return((FUNCTION_RETURN_TYPE)1); 
	} 
 
	/* notify parent that we're up and running */ 
	*(threadInitDataP->threadInitialisedP) = TRUE; 
    threadCondSignal(syncCondVar); 
    threadMutexUnlock(workMutex); 
 
    /* do forever */ 
	messageBufferLength = sizeof(messageBuffer); 
 
	keepLooping = 1; 
	while(keepLooping) 
	{ 
		/* wait forever for next message */ 
		gmo.Options = MQGMO_WAIT + 
		              MQGMO_NO_SYNCPOINT; 
        gmo.WaitInterval = MQWI_UNLIMITED; 
		memcpy(&md, &mdDefault, sizeof(md)); 
 
		memset(messageBuffer, 0, sizeof(messageBuffer)); 
 
		MQGET(hConn,                        /* connection handle */ 
			  requestQHandle,               /* object handle */ 
			  &md,                          /* message descriptor */ 
			  &gmo,                         /* get message options */ 
			  messageBufferLength,          /* size of receive buffer */ 
			  messageBuffer,                /* message buffer */ 
	  (long *)&messageLength,               /* returned message length */ 
			  &compCode,                    /* completion code */ 
			  &reason);                     /* reason code */ 
 
		if (reason != MQRC_NONE) 
		{ 
			printf("Thread %d: MQGET failed with compcode=%d, reason=%d\n", 
			       threadId, 
				   compCode, reason); 
			return((FUNCTION_RETURN_TYPE)1); 
		} 
 
#if 0 
		/* build the reply message by adding 'reply' to end of request msg */ 
		strcat(messageBuffer, " (reply)"); 
#else 
		/* build the XML reply message */ 
		strcpy(messageBuffer, XML_MESSAGE); 
#endif 
		messageLength = strlen(messageBuffer); 
 
		/* print contents of message */ 
		printf("Thread %d: Returned message content is '%s'\n", 
			   threadId, 
			   messageBuffer); 
 
		/* put message on reply to queue */ 
		printf("Thread %d: Requestors reply to queue is '%.*s'\n", threadId, 
			    sizeof(md.ReplyToQ), md.ReplyToQ); 
 
 	    memcpy(&od, &odDefault, sizeof(od)); 
		strncpy(od.ObjectName, 
				md.ReplyToQ, 
				(size_t)MQ_Q_NAME_LENGTH); 
 
		/* set correl id of reply to msg id of request */ 
		/* display the CorrelId contents */
		memcpy(requestId, md.MsgId, MQ_MSG_ID_LENGTH); 
		dispHex((char *)requestId, MQ_MSG_ID_LENGTH, hexBuffer);
		fprintf(stderr, "Request MsgId value is: 0x%s\n", hexBuffer);

		memcpy(&md, &mdDefault, sizeof(md)); 
		memcpy(md.CorrelId, requestId, MQ_CORREL_ID_LENGTH); 
		memcpy(md.ReplyToQ, "Dummy", 6); 
 
	    /* put the message on the queue */ 
		MQPUT1(hConn,               /* connection handle */ 
			   &od,                 /* object descriptor */ 
			   &md,                 /* message descriptor */ 
			   &pmo,                /* put options */ 
			   messageLength,       /* msg length */ 
			   messageBuffer,       /* message buffer */ 
			   &compCode,           /* completion code */ 
			   &reason);            /* reason code */ 
 
		if (reason != MQRC_NONE) 
		{ 
			printf("Thread %d: MQPUT1 failed with compcode=%d, reason=%d\n", threadId, 
				   compCode, reason); 
			return((FUNCTION_RETURN_TYPE)1); 
		} 
	} 
 
	MQDISC(&hConn, 
		   &compCode, 
		   &reason); 
	if (reason != MQRC_NONE) 
	{ 
		printf("Thread %d: MQDISC failed with compcode=%d, reason=%d\n", threadId, 
			   compCode, reason); 
		return((FUNCTION_RETURN_TYPE)1); 
	} 
 
	return(0); 
} 
 
void dispHex(IN char *inputString, 
			 IN int inputStringlength, 
			 OUT char *hexString)
{

	int i;
	char tempBuf[10];
    unsigned char *ch; 				        /* Current Buffer position */

    ch = (unsigned char *)inputString;

	/* start with null-terminated string */
	*hexString = '\0';

	for (i=0; i<inputStringlength; i++)
	{
		sprintf(tempBuf, "%0.2x", *ch++);
		strcat(hexString, tempBuf);
	}
    return;
}
